/**
 * vim: set ts=4 :
 * =============================================================================
 * SourceMod Sample Extension
 * Copyright (C) 2004-2008 AlliedModders LLC.  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, AlliedModders LLC gives you permission to link the
 * code of this program (as well as its derivative works) to "Half-Life 2," the
 * "Source Engine," the "SourcePawn JIT," and any Game MODs that run on software
 * by the Valve Corporation.  You must obey the GNU General Public License in
 * all respects for all other code used.  Additionally, AlliedModders LLC grants
 * this exception to all derivative works.  AlliedModders LLC defines further
 * exceptions, found in LICENSE.txt (as of this writing, version JULY-31-2007),
 * or <http://www.sourcemod.net/license.php>.
 *
 * Version: $Id$
 */

#include "extension.h"
#include "vfuncs.h"
#include "CDetour/detours.h"

/**
 * @file extension.cpp
 * @brief Implement extension code here.
 */

Dukehacks g_Dukehacks;		/**< Global singleton for extension's main interface */

SMEXT_LINK(&g_Dukehacks);
IGameConfig *g_pGameConf = NULL;

/** globals **/
void *g_EntList = NULL;
CUtlVector<HookList> gHookList;
Vector vecOrigin(0.0f, 0.0f, 0.0f);
int gMaxClients = 0;

int g_DamageOffset = 0;

/** SM forwards **/
IForward *g_pEntityCreated = NULL;
IForward *g_pEntitySpawned = NULL;
IForward *g_pEntityDeleted = NULL;

// client hooks
IChangeableForward *g_fwdTraceAttack = NULL;
IChangeableForward *g_fwdProcessUsercmds = NULL;
IChangeableForward *g_fwdPreThink = NULL;
IChangeableForward *g_fwdPostThink = NULL;
IChangeableForward *g_fwdTakeDamage = NULL;

/** hooks **/
SH_DECL_HOOK1_void(IServerGameClients, ClientDisconnect, SH_NOATTRIB, 0, edict_t *);
SH_DECL_HOOK2_void(IServerGameClients, ClientPutInServer, SH_NOATTRIB, 0, edict_t *, char const *);
SH_DECL_HOOK3_void(IServerGameDLL, ServerActivate, SH_NOATTRIB, 0, edict_t *, int, int)

// **** entity hooks ***
#define REDEFINEME 0
// client hooks
SH_DECL_MANUALHOOK3_void(TraceAttackHook, REDEFINEME, 0, 0, CTakeDamageInfoHack &, const Vector &, trace_t *);
SH_DECL_MANUALHOOK5_void(ProcessUsercmdsHook, REDEFINEME, 0, 0, CUserCmd *, int, int, int, bool);
SH_DECL_MANUALHOOK0_void(PreThinkHook, REDEFINEME, 0, 0);
SH_DECL_MANUALHOOK0_void(PostThinkHook, REDEFINEME, 0, 0);
// entity hooks
SH_DECL_MANUALHOOK1_void(TouchHook, REDEFINEME, 0, 0, CBaseEntity *);



/** function pointers **/
AddListenerEntityFunction funcAddListenerEntity = NULL;
TakeDamageFunction funcTakeDamage = NULL;

/** detours **/
CDetour *g_detTakeDamage = NULL;

/** "hack" code **/

int CTakeDamageInfoHack::GetAttacker() const {
	return this->m_hAttacker.GetEntryIndex();
}

int CTakeDamageInfoHack::GetInflictor() const {
	return this->m_hInflictor.GetEntryIndex();
}

void CTakeDamageInfo::Init( CBaseEntity *pInflictor, CBaseEntity *pAttacker, const Vector &damageForce, const Vector &damagePosition, const Vector &reportedPosition, float flDamage, int bitsDamageType, int iCustomDamage )
{
	m_hInflictor = pInflictor;
	if ( pAttacker )
	{
		m_hAttacker = pAttacker;
	}
	else
	{
		m_hAttacker = pInflictor;
	}

	m_flDamage = flDamage;

	m_flBaseDamage = BASEDAMAGE_NOT_SPECIFIED;

	m_bitsDamageType = bitsDamageType;
	m_iDamageCustom = iCustomDamage;

	m_flMaxDamage = flDamage;
	m_vecDamageForce = damageForce;
	m_vecDamagePosition = damagePosition;
	m_vecReportedPosition = reportedPosition;
	m_iAmmoType = -1;
}

CTakeDamageInfo::CTakeDamageInfo()
{
	Init( NULL, NULL, vecOrigin, vecOrigin, vecOrigin, 0, 0, 0 );
}



/** MACROS **/

#define SIGFIND(configname, funcname, functype) \
	if (!g_pGameConf->GetMemSig(configname, &add)) \
	{ \
		g_pSM->LogError(myself, "Could not locate function " #configname); \
		return false; \
	} \
	else \
	{ \
		funcname = (functype)add; \
		g_pSM->LogMessage(myself,"%p " #configname, funcname); \
	}

#define SETUPHOOK(hookname, hookdef) \
	g_pGameConf->GetOffset(hookname, &hookoffset); \
	if (hookoffset>0) \
	{ \
		SH_MANUALHOOK_RECONFIGURE(hookdef, hookoffset, 0, 0); \
		g_pSM->LogMessage(myself, #hookname " offset = %d", hookoffset); \
	} \
	else \
	{ \
		g_pSM->LogError(myself, #hookname " offset not valid"); \
		return false; \
	}


/** detours **/
DETOUR_DECL_MEMBER1_void(TakeDamage, CTakeDamageInfoHack &, info)
{

	CBaseEntity *pEnt = reinterpret_cast<CBaseEntity *>(this);
	int victim = engine->IndexOfEdict(servergameents->BaseEntityToEdict(pEnt));
	int attacker = info.GetAttacker();
	int inflictor = info.GetInflictor();
	int damagetype = info.m_bitsDamageType;
	float damage;



	// print damage flags to console
	Msg("damage bits: ");
	for (int i=0;i<=31;i++)
	{
		if (info.m_bitsDamageType & (1<<i))
		{
			Msg("%d ", i);
		}
	}
	Msg ("\n");

	// print damage flags to console
	Msg("damage custom: ");
	for (int i=0;i<=31;i++)
	{
		if (info.m_iDamageStats & (1<<i))
		{
			Msg("%d ", i);
		}
	}
	Msg ("\n");


	if (victim>0 && victim<gMaxClients && g_DamageOffset>0)
	{
		damage =  info.GetDamage();

		float multiplier = 1.0f;

		if(g_fwdTakeDamage->GetFunctionCount() > 0)
		{
			cell_t result = 0;
			g_fwdTakeDamage->PushCell(victim);
			g_fwdTakeDamage->PushCell(attacker);
			g_fwdTakeDamage->PushCell(inflictor);
			g_fwdTakeDamage->PushFloat(damage);
			g_fwdTakeDamage->PushFloatByRef(&multiplier);
			g_fwdTakeDamage->PushCell(damagetype);
			g_fwdTakeDamage->Execute(&result);

			// block TraceAttack
			if(result==Pl_Handled || result==Pl_Stop)
				return;

			// damage changed
			if (result==Pl_Changed)
			{
				damage *= multiplier;
				info.SetDamage(damage);
				DETOUR_MEMBER_CALL(TakeDamage)(info);
				return;
			}
		}

 
	}
	
	DETOUR_MEMBER_CALL(TakeDamage)(info);
	return;
}


/** class functions **/

bool Dukehacks::SDK_OnLoad(char *error, size_t maxlength, bool late)
{

	/* load game configuration file */
	char conf_error[255];
	if (!gameconfs->LoadGameConfigFile(GAMECONFIGFILE, &g_pGameConf, conf_error, sizeof(conf_error)))
	{
		if (conf_error)
		{
			snprintf(error, maxlength, "Could not read %s.txt: %s", GAMECONFIGFILE, conf_error);
		}
		return false;
	}

	/* find function signatures */
	void *add;
	SIGFIND("AddListenerEntity", funcAddListenerEntity, AddListenerEntityFunction);
	SIGFIND("TakeDamage", funcTakeDamage, TakeDamageFunction);

	char *addr = NULL;
#ifdef WIN32
	int offset;
	/* gEntList and/or g_pEntityList */
	if (!g_pGameConf->GetMemSig("LevelShutdown", (void **)&addr) || !addr)
	{
		g_pSM->LogError(myself, "Could not locate LevelShutdown for gEntList");
		return false;
	}
	if (!g_pGameConf->GetOffset("gEntList", &offset) || !offset)
	{
		g_pSM->LogError(myself, "Could not locate gEntList from offset (WINDOWS)");
		return false;
	}
	g_EntList = *reinterpret_cast<void **>(addr + offset);
	g_pSM->LogMessage(myself,"%p gEntList", g_EntList);

	
#else
	/* gEntList and/or g_pEntityList */
	if (!g_pGameConf->GetMemSig("gEntList", (void **)&addr) || !addr)
	{
		g_pSM->LogError(myself, "Could not load gEntList (LINUX)");
		return false;
	}
	g_EntList = reinterpret_cast<void *>(addr);
	g_pSM->LogMessage(myself,"%p gEntList", g_EntList);

#endif

	/* setup hooks based on offsets in file */
	int hookoffset;
	
	SETUPHOOK("TraceAttack", TraceAttackHook);
	SETUPHOOK("ProcessUsercmds", ProcessUsercmdsHook);
	SETUPHOOK("PreThink", PreThinkHook);
	SETUPHOOK("PostThink", PostThinkHook);

	SETUPHOOK("Touch", TouchHook);

	/* load other game config info */
	g_pGameConf->GetOffset("CTakeDamageInfo.m_flDamage", &g_DamageOffset);
	g_pSM->LogMessage(myself, "CTakeDamageInfo.m_flDamage offset = %d", g_DamageOffset); \

	/* vfuncs */
	InitializeVFuncs(g_pGameConf);

	/* listen for entity create/spawn/delete */
	scanAddListenerEntity(this);

	/* setup detours */
	CDetourManager::Init(g_pSM->GetScriptingEngine(), g_pGameConf);

	g_detTakeDamage = DETOUR_CREATE_MEMBER(TakeDamage, "TakeDamage");
	if (g_detTakeDamage != NULL)
	{
		g_detTakeDamage->EnableDetour();
	}

	return true;
}

void Dukehacks::SDK_OnUnload()
{
	forwards->ReleaseForward(g_pEntityCreated);
	forwards->ReleaseForward(g_pEntitySpawned);
	forwards->ReleaseForward(g_pEntityDeleted);

	// client hooks
	forwards->ReleaseForward(g_fwdTraceAttack);
	forwards->ReleaseForward(g_fwdProcessUsercmds);
	forwards->ReleaseForward(g_fwdPreThink);
	forwards->ReleaseForward(g_fwdPostThink);
	forwards->ReleaseForward(g_fwdTakeDamage);

}

void Dukehacks::SDK_OnAllLoaded()
{
	sharesys->AddNatives(myself, MyNatives);
	ParamType p1[] = {Param_Cell};
	
	g_pEntityCreated = forwards->CreateForward("dhOnEntityCreated", ET_Ignore, 1, p1);
	g_pEntitySpawned = forwards->CreateForward("dhOnEntitySpawned", ET_Ignore, 1, p1);
	g_pEntityDeleted = forwards->CreateForward("dhOnEntityDeleted", ET_Ignore, 1, p1);
	
	g_fwdTraceAttack = forwards->CreateForwardEx(NULL, ET_Hook, 5, NULL, Param_Cell, Param_Cell, Param_Cell, Param_Float, Param_FloatByRef);
	g_fwdProcessUsercmds = forwards->CreateForwardEx(NULL, ET_Hook, 3, NULL, Param_Cell, Param_CellByRef, Param_CellByRef);
	g_fwdPreThink = forwards->CreateForwardEx(NULL, ET_Hook, 1, NULL, Param_Cell);
	g_fwdPostThink = forwards->CreateForwardEx(NULL, ET_Hook, 1, NULL, Param_Cell);
	g_fwdTakeDamage = forwards->CreateForwardEx(NULL, ET_Hook, 6, NULL, Param_Cell, Param_Cell, Param_Cell, Param_Float, Param_FloatByRef, Param_Cell);
}

bool Dukehacks::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlength, bool late)
{
	size_t maxlen = maxlength;

	// add standard plugin hooks
	SH_ADD_HOOK_MEMFUNC(IServerGameClients, ClientDisconnect, serverclients, &g_Dukehacks, &Dukehacks::ClientDisconnect, true);
	SH_ADD_HOOK_MEMFUNC(IServerGameClients, ClientPutInServer, serverclients, &g_Dukehacks, &Dukehacks::ClientPutInServer, true);
	SH_ADD_HOOK_MEMFUNC(IServerGameDLL, ServerActivate, gamedll, &g_Dukehacks, &Dukehacks::ServerActivate, true);
	return true;
}

bool Dukehacks::SDK_OnMetamodUnload(char *error, size_t maxlength)
{
	// remove standard plugin hooks
	SH_REMOVE_HOOK_MEMFUNC(IServerGameClients, ClientDisconnect, serverclients, &g_Dukehacks, &Dukehacks::ClientDisconnect, true);
	SH_REMOVE_HOOK_MEMFUNC(IServerGameClients, ClientPutInServer, serverclients, &g_Dukehacks, &Dukehacks::ClientPutInServer, true);
	SH_REMOVE_HOOK_MEMFUNC(IServerGameDLL, ServerActivate, gamedll, &g_Dukehacks, &Dukehacks::ServerActivate, true);
	return true;
}

// *************************************************
// Standard MM:S Plugin Hooks
void Dukehacks::ServerActivate(edict_t *pEdictList, int edictCount, int clientMax)
{
	gMaxClients = clientMax;
	RETURN_META(MRES_IGNORED);
}

void Dukehacks::ClientPutInServer(edict_t *pEntity, char const *playername)
{
	// add player hooks
	IServerUnknown *unk = pEntity->GetUnknown();
	if (unk)
	{
		CBaseEntity *pBase = unk->GetBaseEntity();
		if (pBase)
		{
			SH_ADD_MANUALHOOK_MEMFUNC(TraceAttackHook, pBase, &g_Dukehacks, &Dukehacks::TraceAttack_handler, false); 
			SH_ADD_MANUALHOOK_MEMFUNC(ProcessUsercmdsHook, pBase, &g_Dukehacks, &Dukehacks::ProcessUsercmds_handler, false); 
			SH_ADD_MANUALHOOK_MEMFUNC(PreThinkHook, pBase, &g_Dukehacks, &Dukehacks::PreThink_handler, false); 
			SH_ADD_MANUALHOOK_MEMFUNC(PostThinkHook, pBase, &g_Dukehacks, &Dukehacks::PostThink_handler, false); 
		}
	}
	RETURN_META(MRES_IGNORED);
}

void Dukehacks::ClientDisconnect(edict_t *pEntity)
{
	// remove player hooks
	IServerUnknown *unk = pEntity->GetUnknown();
	if (unk)
	{
		CBaseEntity *pBase = unk->GetBaseEntity();
		if (pBase)
		{
			SH_REMOVE_MANUALHOOK_MEMFUNC(TraceAttackHook, pBase, &g_Dukehacks, &Dukehacks::TraceAttack_handler, false); 
			SH_REMOVE_MANUALHOOK_MEMFUNC(ProcessUsercmdsHook, pBase, &g_Dukehacks, &Dukehacks::ProcessUsercmds_handler, false); 
			SH_REMOVE_MANUALHOOK_MEMFUNC(PreThinkHook, pBase, &g_Dukehacks, &Dukehacks::PreThink_handler, false);
			SH_REMOVE_MANUALHOOK_MEMFUNC(PostThinkHook, pBase, &g_Dukehacks, &Dukehacks::PostThink_handler, false); 
		}
	}
	RETURN_META(MRES_IGNORED);
}

// *************************************************
// Entity Listener Functions

void Dukehacks::OnEntityCreated( CBaseEntity *pEntity )
{
	if (!g_pEntityCreated)
		return;
	
	SendListenerForward(g_pEntityCreated, pEntity);
}

void Dukehacks::OnEntitySpawned( CBaseEntity *pEntity )
{
	if (!g_pEntitySpawned)
		return;

	SendListenerForward(g_pEntitySpawned, pEntity);	
}

void Dukehacks::OnEntityDeleted( CBaseEntity *pEntity )
{
	edict_t *pEnt = servergameents->BaseEntityToEdict(pEntity);
	int entity = engine->IndexOfEdict( pEnt );
	
	if (entity > 1)
	{
		int hookcount = gHookList.Count();
		int i;
		for (i=hookcount-1; i>=0; i--)
		{
			if (gHookList[i].entityid == entity)
			{
				UnHookEntity(pEntity, gHookList[i].type, i);
			}
		}
	}

	if (!g_pEntityDeleted)
		return;

	SendListenerForward(g_pEntityDeleted, pEntity);
}


void Dukehacks::SendListenerForward(IForward *SMForward, CBaseEntity *pEntity)
{
	int index = -1;
	edict_t *pEnt = servergameents->BaseEntityToEdict(pEntity);
	if (pEnt && !pEnt->IsFree())
	{
		index = engine->IndexOfEdict( pEnt );
	}
	if (index > 1)
	{
		// forward to SM plugins
		cell_t entindex = index;
		cell_t result = 0;
		if (SMForward->GetFunctionCount())
		{
			SMForward->PushCell(entindex);
			SMForward->Execute(&result);
		}
	}
}

// *************************************************
// Hook handlers

void Dukehacks::TraceAttack_handler(CTakeDamageInfoHack &info, const Vector &vecDir, trace_t *ptr)
{
	CBaseEntity *pEnt = META_IFACEPTR(CBaseEntity);
	int victim = engine->IndexOfEdict(servergameents->BaseEntityToEdict(pEnt));
	int attacker = info.GetAttacker();
	int inflictor = info.GetInflictor();
	float damage;

	if (g_DamageOffset>0)
	{
		damage = info.GetDamage();

		float multiplier = 1.0f;

		if(g_fwdTraceAttack->GetFunctionCount() > 0)
		{
			cell_t result = 0;
			g_fwdTraceAttack->PushCell(victim);
			g_fwdTraceAttack->PushCell(attacker);
			g_fwdTraceAttack->PushCell(inflictor);
			g_fwdTraceAttack->PushFloat(damage);
			g_fwdTraceAttack->PushFloatByRef(&multiplier);
			g_fwdTraceAttack->Execute(&result);

			// block TraceAttack
			if(result==Pl_Handled || result==Pl_Stop)
				RETURN_META(MRES_SUPERCEDE);

			// damage changed
			if (result==Pl_Changed)
			{
				damage *= multiplier;
				info.SetDamage(damage);
			}
		}

	}

	/*
	//use to find offset if it changes
	void *add = &info;
	float temp;
	int temp2;
	for (int Offset=0; Offset<68; Offset+=4)
	{
		temp = *(float *)((char *)add + Offset );
		temp2 = *(int *)((char *)add + Offset );
		g_pSM->LogMessage(myself, "offset: %d  float: %f int: %d", Offset, temp, temp2);
	}
	*/
	

	RETURN_META(MRES_IGNORED);
}

void Dukehacks::ProcessUsercmds_handler( CUserCmd *cmds, int numcmds, int totalcmds, int dropped_packets, bool paused )
{

	static int count = 0;
	CBaseEntity *pBase = META_IFACEPTR(CBaseEntity);
	edict_t *pEnt = pBase->edict();

	int client = engine->IndexOfEdict(pEnt);
	int buttons = cmds->buttons;
	int impulse = cmds->impulse;

	if(g_fwdProcessUsercmds->GetFunctionCount() > 0)
	{
		cell_t result = 0;
		g_fwdProcessUsercmds->PushCell(client);
		g_fwdProcessUsercmds->PushCellByRef(&buttons);
		g_fwdProcessUsercmds->PushCellByRef(&impulse);
		g_fwdProcessUsercmds->Execute(&result);

		// block ProcessUsercmds
		if(result==Pl_Handled || result==Pl_Stop)
			RETURN_META(MRES_SUPERCEDE);

		// changes made
		if (result==Pl_Changed)
		{
			cmds->buttons = buttons;
			cmds->impulse = impulse;
			RETURN_META(MRES_HANDLED);
		}
	}
	RETURN_META(MRES_IGNORED);
}

void Dukehacks::PreThink_handler( void )
{

	static int count = 0;
	CBaseEntity *pBase = META_IFACEPTR(CBaseEntity);
	edict_t *pEnt = pBase->edict();
	int client = engine->IndexOfEdict(pEnt);
	
	if(g_fwdPreThink->GetFunctionCount() > 0)
	{
		cell_t result = 0;
		g_fwdPreThink->PushCell(client);
		g_fwdPreThink->Execute(&result);
	}

	RETURN_META(MRES_IGNORED);

}

void Dukehacks::PostThink_handler( void )
{

	static int count = 0;
	CBaseEntity *pBase = META_IFACEPTR(CBaseEntity);
	edict_t *pEnt = pBase->edict();
	int client = engine->IndexOfEdict(pEnt);
	
	if(g_fwdPostThink->GetFunctionCount() > 0)
	{
		cell_t result = 0;
		g_fwdPostThink->PushCell(client);
		g_fwdPostThink->Execute(&result);
	}

	RETURN_META(MRES_IGNORED);

}


void Dukehacks::Touch_handler(CBaseEntity *pOther)
{
	static IPluginFunction *func = NULL;

	CBaseEntity *pEnt = META_IFACEPTR(CBaseEntity);
	int entity = engine->IndexOfEdict(servergameents->BaseEntityToEdict(pEnt));
	int other = engine->IndexOfEdict(servergameents->BaseEntityToEdict(pOther));

	cell_t result = 0;
	cell_t hires = Pl_Continue;

	int hookcount = gHookList.Count();
	int i;
	for (i=hookcount-1; i>=0; i--)
	{
		if (gHookList[i].entityid == entity)
		{
			if (gHookList[i].type == EHK_Touch)
			{
				func = gHookList[i].func;
				
				func->PushCell(entity);
				func->PushCell(other);
				func->Execute(&result);
				if (result>hires)
					hires = result;
			}
		}
	}
	
	// block Touch
	if(hires==Pl_Handled || hires==Pl_Stop)
		RETURN_META(MRES_SUPERCEDE);

	// no change
	RETURN_META(MRES_IGNORED);
}

// *************************************************
// natives
cell_t dhAddClientHook(IPluginContext *pContext, const cell_t *params)
{
	if(params[0] != 2)
	{
		pContext->ThrowNativeError("native dhAddClientHook must use 2 args...");
		return false;
	}

	int hooktype = params[1];
	funcid_t funcid = static_cast<funcid_t>(params[2]);

	switch(hooktype)
	{
		case CHK_TraceAttack:
			g_fwdTraceAttack->AddFunction(pContext, funcid);
			return true;
			break;

		case CHK_ProcessUsercmds:
			g_fwdProcessUsercmds->AddFunction(pContext, funcid);
			return true;
			break;

		case CHK_PreThink:
			g_fwdPreThink->AddFunction(pContext, funcid);
			return true;
			break;

		case CHK_PostThink:
			g_fwdPostThink->AddFunction(pContext, funcid);
			return true;
			break;

		case CHK_TakeDamage:
			g_fwdTakeDamage->AddFunction(pContext, funcid);
			return true;
			break;

		default:
			// nothing yet
			break;
	}

	return true;
}

cell_t dhHookEntity(IPluginContext *pContext, const cell_t *params)
{
	int entid = (int)params[1];
	EntityHookType hooktype = (EntityHookType)params[2];
	IPluginFunction *func = pContext->GetFunctionById(params[3]);

	edict_t *pEnt = engine->PEntityOfEntIndex(entid);
	if (!pEnt || pEnt->IsFree())
	{
		return false;
	}

	CBaseEntity *pBase = servergameents->EdictToBaseEntity(pEnt);
	if (!pBase)
	{
		return false;
	}

	switch (hooktype)
	{
	case (EHK_Touch):
			SH_ADD_MANUALHOOK_MEMFUNC(TouchHook, pBase, &g_Dukehacks, &Dukehacks::Touch_handler, false);
			HookList newhook;
			newhook.entityid = entid;
			newhook.type = hooktype;
			newhook.func = func;
			gHookList.AddToTail(newhook);
			return true;

	default:
			return false;
	}
	
	return false;
}

cell_t dhUnHookEntity(IPluginContext *pContext, const cell_t *params)
{	
	int entity = (int)params[1];
	EntityHookType hooktype = (EntityHookType)params[2];

	if (entity > 1)
	{
		int hookcount = gHookList.Count();
		int i;
		for (i=hookcount-1; i>=0; i--)
		{
			if (gHookList[i].entityid == entity)
			{
				if (gHookList[i].type = hooktype)
				{
					edict_t *pEntity = engine->PEntityOfEntIndex(entity);
					if (pEntity && !pEntity->IsFree())
					{
						IServerUnknown *unk = pEntity->GetUnknown();
						if (unk)
						{
							CBaseEntity *pBase = unk->GetBaseEntity();
							if (pBase)
							{
								return UnHookEntity(pBase, gHookList[i].type, i);
							}
						}
					}
				}
			}
		}
	}
	
	return false;
}

cell_t dhTakeDamage(IPluginContext *pContext, const cell_t *params)
{
	int entity = (int)params[1];
	int attacker = (int)params[2];
	int inflictor = (int)params[3];
	float damage = sp_ctof(params[4]);
	int damagetype = (int)params[5]; 

	// get entity
	edict_t *pEntity = NULL;
	pEntity = engine->PEntityOfEntIndex(entity);
	if (!pEntity || pEntity->IsFree())
		return false;
	CBaseEntity *pEnt = servergameents->EdictToBaseEntity(pEntity);
	if (!pEnt)
		return false;

	// get attacker
	pEntity = NULL;
	pEntity = engine->PEntityOfEntIndex(entity);
	if (!pEntity || pEntity->IsFree())
		return false;
	CBaseEntity *pAtt = servergameents->EdictToBaseEntity(pEntity);
	if (!pEnt)
		return false;

	// get inflictor
	pEntity = NULL;
	pEntity = engine->PEntityOfEntIndex(entity);
	if (!pEntity || pEntity->IsFree())
		return false;
	CBaseEntity *pInf = servergameents->EdictToBaseEntity(pEntity);
	if (!pEnt)
		return false;

	// call scanned TakeDamage function
	CTakeDamageInfoHack i2;
	i2.Init(pInf, pAtt, vecOrigin, vecOrigin, vecOrigin, damage, damagetype, -1); 
	scanTakeDamage(pEnt, i2);

	return true;
}

// *************************************************
// helpers
bool UnHookEntity(CBaseEntity *pBase, EntityHookType HookType, int HookListIdx)
{
	switch (HookType)
	{
	case (EHK_Touch):
			SH_REMOVE_MANUALHOOK_MEMFUNC(TouchHook, pBase, &g_Dukehacks, &Dukehacks::Touch_handler, false);
			gHookList.Remove(HookListIdx);
			return true;

	default:
			return false;
	}
}

// *************************************************
// Scanned functions
void scanAddListenerEntity( IEntityListener *pListener )
{
	#ifdef WIN32
		funcAddListenerEntity(g_EntList /*this*/, 0 /*dummy*/, pListener);
	#else
		(funcAddListenerEntity)(g_EntList, pListener);
	#endif
}

void scanTakeDamage( CBaseEntity *pThis, const CTakeDamageInfoHack &info )
{
	#ifdef WIN32
		funcTakeDamage(pThis /*this*/, 0 /*dummy*/, info);
	#else
		(funcTakeDamage)(pThis, info);
	#endif
}

