/**
 * vim: set ts=4 :
 * =============================================================================
 * SourceMod Sample Extension
 * Copyright (C) 2004-2008 AlliedModders LLC.  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, AlliedModders LLC gives you permission to link the
 * code of this program (as well as its derivative works) to "Half-Life 2," the
 * "Source Engine," the "SourcePawn JIT," and any Game MODs that run on software
 * by the Valve Corporation.  You must obey the GNU General Public License in
 * all respects for all other code used.  Additionally, AlliedModders LLC grants
 * this exception to all derivative works.  AlliedModders LLC defines further
 * exceptions, found in LICENSE.txt (as of this writing, version JULY-31-2007),
 * or <http://www.sourcemod.net/license.php>.
 *
 * Version: $Id$
 */

#ifndef _INCLUDE_SOURCEMOD_EXTENSION_PROPER_H_
#define _INCLUDE_SOURCEMOD_EXTENSION_PROPER_H_

/**
 * @file extension.h
 * @brief dukehacks extension
 */

#include "smsdk_ext.h"
#include "defines.h"
#define GAME_DLL 1
#include "cbase.h"
#include "takedamageinfo.h"

// *************************************************
// globals
class HookList {
public:
	bool operator==(const HookList& rhs) const { return ( (rhs.entityid == this->entityid) && (rhs.type == this->type)); } ;
	int entityid;
	EntityHookType type;
	IPluginFunction *func;
};
extern CUtlVector<HookList> gHookList;

// *************************************************
// natives
cell_t dhAddClientHook(IPluginContext *pContext, const cell_t *params);
cell_t dhHookEntity(IPluginContext *pContext, const cell_t *params);
cell_t dhUnHookEntity(IPluginContext *pContext, const cell_t *params);
cell_t dhTakeDamage(IPluginContext *pContext, const cell_t *params);

// *************************************************
// helpers
bool UnHookEntity(CBaseEntity *pBase, EntityHookType HookType, int HookListIdx);



const sp_nativeinfo_t MyNatives[] = 
{
	{"dhAddClientHook", dhAddClientHook},
	{"dhHookEntity", dhHookEntity},
	{"dhUnHookEntity", dhUnHookEntity},
	{"dhTakeDamage", dhTakeDamage},
	{NULL,			NULL}
};

class CTakeDamageInfoHack : public CTakeDamageInfo {
public:
	int GetAttacker() const;
	int GetInflictor() const;
	inline float GetDamage() const { return m_flDamage; };
	inline void SetDamage( float flDamage ) { m_flDamage = flDamage; };
};


/**
 * @brief Dukehacks extension
 */
class Dukehacks : public SDKExtension, public IEntityListener
{
private:


public:

	virtual bool SDK_OnLoad(char *error, size_t maxlength, bool late);
	virtual void SDK_OnUnload();
	virtual void SDK_OnAllLoaded();
	//virtual void SDK_OnPauseChange(bool paused);
	//virtual bool QueryRunning(char *error, size_t maxlength);


	// Standard MM:S Plugin Hooks
	void ServerActivate(edict_t *pEdictList, int edictCount, int clientMax);
	void ClientPutInServer(edict_t *pEntity, char const *playername);
	void ClientDisconnect(edict_t *pEntity);

	// Entity Listener
	virtual void OnEntityCreated( CBaseEntity *pEntity );
	virtual void OnEntitySpawned( CBaseEntity *pEntity );
	virtual void OnEntityDeleted( CBaseEntity *pEntity );
	void SendListenerForward(IForward *SMForward, CBaseEntity *pEntity);

	// hook handlers
	void TraceAttack_handler(CTakeDamageInfoHack &info, const Vector &vecDir, trace_t *ptr);
	void ProcessUsercmds_handler( CUserCmd *cmds, int numcmds, int totalcmds, int dropped_packets, bool paused );
	void PreThink_handler( void );
	void PostThink_handler( void );
	void OnDamagedByExplosion_handler(CTakeDamageInfoHack &info);
	void OnTakeDamage_handler(CTakeDamageInfoHack &info);

	void Touch_handler(CBaseEntity *pOther);

public:
#if defined SMEXT_CONF_METAMOD

	virtual bool SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlength, bool late);
	virtual bool SDK_OnMetamodUnload(char *error, size_t maxlength);
	//virtual bool SDK_OnMetamodPauseChange(bool paused, char *error, size_t maxlength);

#endif
};

// *************************************************
// Scanned functions

#ifdef WIN32
	typedef void (__fastcall *AddListenerEntityFunction)(void* /*this*/, int /*dummy*/, IEntityListener *);
	typedef void (__fastcall *TakeDamageFunction)(void* /*this*/, int /*dummy*/, const CTakeDamageInfoHack &);
#else
	typedef void (*AddListenerEntityFunction)(void*, IEntityListener *);
	typedef void (*TakeDamageFunction)(void*, const CTakeDamageInfoHack &);
#endif


void scanAddListenerEntity( IEntityListener *pListener );
void scanTakeDamage( CBaseEntity *pThis, const CTakeDamageInfoHack &info );



#endif // _INCLUDE_SOURCEMOD_EXTENSION_PROPER_H_
